# -*- coding: utf-8 -*-

"""A wrapper around ROBOT functionality.

.. seealso:: https://robot.obolibrary.org
"""

import dataclasses
import json
import logging
import os
import subprocess
import tempfile
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from subprocess import check_output
from typing import List, Literal, Optional, Union

import bioregistry
import requests

from .obograph import Graph, GraphDocument

__all__ = [
    "ParseResults",
    "convert_to_obograph_local",
    "convert_to_obograph_remote",
    "convert_to_obograph",
    "get_obograph_by_prefix",
]

logger = logging.getLogger(__name__)


@dataclass
class ParseResults:
    """A dataclass containing an OBO Graph JSON and text output from ROBOT."""

    graph_document: GraphDocument
    messages: List[str] = dataclasses.field(default_factory=list)

    def squeeze(self) -> Graph:
        """Get the first graph."""
        return self.graph_document.graphs[0]


def get_obograph_by_prefix(
    prefix: str, *, json_path: Union[None, str, Path] = None
) -> ParseResults:
    """Get an ontology by its Bioregistry prefix."""
    if prefix != bioregistry.normalize_prefix(prefix):
        raise ValueError("this function requires bioregistry canonical prefixes")

    json_iri = bioregistry.get_json_download(prefix)

    if json_iri is not None:
        res_json = requests.get(json_iri).json()
        graph_document = GraphDocument(**res_json)
        return ParseResults(graph_document=graph_document)

    owl_iri = bioregistry.get_owl_download(prefix)
    obo_iri = bioregistry.get_obo_download(prefix)

    for iri in [owl_iri, obo_iri]:
        if iri is None:
            continue

        try:
            parse_results = convert_to_obograph_remote(iri, json_path=json_path)
        except subprocess.CalledProcessError:
            logger.info("could not parse OBO for %s from %s", prefix, iri)
            continue
        else:
            return parse_results

    raise RuntimeError(f"no IRI available for Bioregistry prefix {prefix}")


def convert_to_obograph_local(
    path: Union[str, Path],
    *,
    json_path: Union[None, str, Path] = None,
) -> ParseResults:
    """Convert a local OWL/OBO file to an OBO Graph JSON object.

    :param path: The path to a local OWL or OBO file
    :param json_path: The optional path to store the intermediate
        OBO Graph JSON file generated by ROBOT. If not given, the
        OBO Graph JSON file will be put in a temporary directory
        and deleted after the function finishes.
    :returns: An object with the parsed OBO Graph JSON and text
        output from the ROBOT conversion program
    """
    return convert_to_obograph(input_str=path, flag="-i", json_path=json_path)


def convert_to_obograph_remote(
    iri: str,
    *,
    json_path: Union[None, str, Path] = None,
) -> ParseResults:
    """Convert a remote OWL/OBO file to an OBO Graph JSON object.

    :param iri: The IRI for a remote OWL or OBO file
    :param json_path: The optional path to store the intermediate
        OBO Graph JSON file generated by ROBOT. If not given, the
        OBO Graph JSON file will be put in a temporary directory
        and deleted after the function finishes.
    :returns: An object with the parsed OBO Graph JSON and text
        output from the ROBOT conversion program
    """
    return convert_to_obograph(input_str=iri, flag="-I", json_path=json_path)


def convert_to_obograph(
    input_str: Union[str, Path],
    *,
    flag: Optional[Literal["-i", "-I"]] = None,
    json_path: Union[None, str, Path] = None,
) -> ParseResults:
    """Convert a local OWL file to a JSON file.

    :param input_str: Either a local file path or IRI. If a local file path
        is used, pass ``"-i"`` to ``flag``. If an IRI is used, pass ``"-I"``
        to ``flag``.
    :param flag: The flag to denote if the file is local or remote.
        Tries to infer from input string if none is given
    :param json_path: The optional path to store the intermediate
        OBO Graph JSON file generated by ROBOT. If not given, the
        OBO Graph JSON file will be put in a temporary directory
        and deleted after the function finishes.
    :returns: An object with the parsed OBO Graph JSON and text
        output from the ROBOT conversion program
    """
    if flag is None:
        flag = "-I" if _is_remote(input_str) else "-i"
    with _path_context(json_path) as path:
        args = ["robot", "convert", flag, str(input_str), "-o", str(path), "--format", "json"]
        ret = check_output(args, cwd=os.path.dirname(__file__))  # noqa:S603
        messages = ret.decode().strip().splitlines()
        graph_document_raw = json.loads(path.read_text())
        for i, graph in enumerate(graph_document_raw["graphs"]):
            if "id" not in graph:
                raise ValueError(f"{input_str} graph [{i}] missing id")
        graph_document = GraphDocument(**graph_document_raw)
        return ParseResults(graph_document=graph_document, messages=messages)


#: Prefixes that denote remote resources
PROTOCOLS = {
    "https://",
    "http://",
    "ftp://",
    "ftps://",
}


def _is_remote(url: Union[str, Path]) -> bool:
    return isinstance(url, str) and any(url.startswith(protocol) for protocol in PROTOCOLS)


@contextmanager
def _path_context(path: Union[None, str, Path], name: str = "output.json"):
    if path is not None:
        yield Path(path).resolve()
    else:
        with tempfile.TemporaryDirectory() as directory:
            yield Path(directory).joinpath(name)



# -*- coding: utf-8 -*-

"""Test API functions."""

import unittest

from bioontologies.robot import ParseResults, convert_to_obograph_remote


class TestAPI(unittest.TestCase):
    """Test parsing a remote file."""

    def test_parse_owl(self):
        """Test parsing a remote JSON graph, should take less than a minute."""
        uri = "https://raw.githubusercontent.com/pato-ontology/pato/master/pato.owl"
        result = convert_to_obograph_remote(uri)
        self.assertIsInstance(result, ParseResults)
        graph = result.squeeze()
        self.assertEqual("PATO - the Phenotype And Trait Ontology", graph.title)
        self.assertEqual("quality", graph.default_namespace)
        self.assertEqual("http://purl.obolibrary.org/obo/PATO_0000001", graph.root)



# -*- coding: utf-8 -*-

"""Tools for biomedical ontologies."""

from .robot import convert_to_obograph, get_obograph_by_prefix

__all__ = [
    "convert_to_obograph",
    "get_obograph_by_prefix",
]



"""Data structures for representing OBO Graphs.

.. seealso:: https://github.com/geneontology/obographs
"""

from collections import defaultdict
from operator import attrgetter
from typing import Any, Iterable, List, Literal, Mapping, Optional, Set, Union

from bioregistry import normalize_curie
from pydantic import BaseModel

__all__ = [
    "Property",
    "Definition",
    "Xref",
    "Synonym",
    "Meta",
    "Edge",
    "Node",
    "Graph",
    "GraphDocument",
]

OBO_URI_PREFIX = "http://purl.obolibrary.org/obo/"


class Property(BaseModel):
    """Represent a property inside a metadata element."""

    pred: str
    val: str


class Definition(BaseModel):
    """Represents a definition for a node."""

    val: str
    # Just a list of CURIEs/IRIs
    xrefs: Optional[List[str]]


class Xref(BaseModel):
    """Represents a cross-reference."""

    val: str


class Synonym(BaseModel):
    """Represents a synonym inside an object meta."""

    pred: str
    val: str
    # Just a list of CURIEs/IRIs
    xrefs: List[str]
    synonymType: Optional[str]  # noqa:N815


class Meta(BaseModel):
    """Represents the metadata about a node or ontology."""

    definition: Optional[Definition]
    subsets: Optional[List[str]]
    xrefs: Optional[List[Xref]]
    synonyms: Optional[List[Synonym]]
    comments: Optional[List]
    version: Optional[str]
    basicPropertyValues: Optional[List[Property]]  # noqa:N815
    deprecated: bool = False


class Edge(BaseModel):
    """Represents an edge in an OBO Graph."""

    sub: str
    pred: str
    obj: str
    meta: Optional[Meta]


class Node(BaseModel):
    """Represents a node in an OBO Graph."""

    id: str
    lbl: Optional[str]
    meta: Optional[Meta]
    type: Optional[Literal["CLASS", "PROPERTY", "INDIVIDUAL"]]
    alternative_ids: Optional[List[str]]

    @property
    def deprecated(self) -> bool:
        """Get if the node is deprecated."""
        if self.meta is None:
            return False
        return self.meta.deprecated

    @property
    def synonyms(self) -> List[Synonym]:
        """Get the synonyms for the node."""
        if self.meta and self.meta.synonyms:
            return self.meta.synonyms
        return []

    @property
    def xrefs(self) -> List[Xref]:
        """Get the xrefs for the node."""
        if self.meta and self.meta.xrefs:
            return self.meta.xrefs
        return []

    @property
    def replaced_by(self) -> Optional[str]:
        """Get the identifier that this node was replaced by."""
        if not self.meta:
            return None
        preds = ["http://purl.obolibrary.org/obo/IAO_0100001", "IAO:0100001", "iao:0100001"]
        for prop in self.meta.basicPropertyValues or []:
            if any(prop.pred == pred for pred in preds):
                return prop.val
        return None


class Graph(BaseModel):
    """A graph corresponds to an ontology."""

    id: str
    meta: Meta
    nodes: List[Node]
    edges: List[Edge]
    equivalentNodesSets: Any  # noqa:N815
    logicalDefinitionAxioms: Any  # noqa:N815
    domainRangeAxioms: Any  # noqa:N815
    propertyChainAxioms: Any  # noqa:N815

    @property
    def roots(self) -> List[str]:
        """Get the ontology root terms."""
        return self._get_properties(
            [
                "http://purl.obolibrary.org/obo/IAO_0000700",
                "IAO:0000700",
            ]
        )

    @property
    def license(self):
        """Get the license of the ontology."""
        return self._get_property("http://purl.org/dc/terms/license")

    @property
    def title(self):
        """Get the title of the ontology."""
        return self._get_property("http://purl.org/dc/elements/1.1/title")

    @property
    def description(self):
        """Get the license of the ontology."""
        return self._get_property("http://purl.org/dc/elements/1.1/description")

    @property
    def version_iri(self) -> Optional[str]:
        """Get the version of the ontology."""
        return self.meta.version

    @property
    def version(self):
        """Get the version of the ontology."""
        return self._get_property("http://www.w3.org/2002/07/owl#versionInfo")

    @property
    def default_namespace(self):
        """Get the version of the ontology."""
        return self._get_property("http://www.geneontology.org/formats/oboInOwl#default-namespace")

    def _get_property(self, pred: Union[str, List[str]]) -> Optional[str]:
        p = self._get_properties(pred)
        return p[0] if p else None

    def _get_properties(self, pred: Union[str, List[str]]) -> Optional[str]:
        if isinstance(pred, str):
            pred = [pred]
        return [
            prop.val
            for prop in self.meta.basicPropertyValues or []
            if any(prop.pred == p for p in pred)
        ]

    def standardize(self, keep_invalid: bool = False) -> "Graph":
        """Standardize the OBO graph.

        :param keep_invalid: Should CURIEs/IRIs that aren't handled
            by the Bioregistry be kept? Defaults to false.
        :returns: This OBO graph, modified in place as follows:

            1. Convert IRIs to CURIEs (in many places) using :mod:`bioregistry`
            2. Add alternative identifiers to :class:`Node` objects
        """
        # Convert URIs to CURIEs
        for node in self.nodes:
            if node.id.startswith(OBO_URI_PREFIX):
                node.id = _clean_uri(node.id, keep_invalid=True)  # type:ignore
            if node.meta:
                for prop in node.meta.basicPropertyValues or []:
                    prop.pred = _clean_uri(prop.pred, keep_invalid=True)  # type:ignore
                    prop.val = _clean_uri(prop.val, keep_invalid=True)  # type:ignore

                for synonym in node.meta.synonyms or []:
                    synonym.pred = _clean_uri(synonym.pred, keep_invalid=True)  # type:ignore
                    if synonym.synonymType:
                        synonym.synonymType = _clean_uri(
                            synonym.synonymType, keep_invalid=True
                        )  # type:ignore

                # Remove self-xrefs, duplicate xrefs
                xrefs: List[Xref] = []
                xrefs_vals: Set[str] = set()
                for xref in node.meta.xrefs or []:
                    if xref.val == node.id:
                        continue
                    new_xref_val = _clean_uri(xref.val, keep_invalid=keep_invalid)
                    if new_xref_val is None:
                        continue
                    xref.val = new_xref_val
                    if xref.val == node.id or xref.val in xrefs_vals:
                        continue
                    xrefs_vals.add(xref.val)
                    xrefs.append(xref)
                node.meta.xrefs = sorted(xrefs, key=attrgetter("val"))

        # TODO add xrefs from definition into node if the are "real" CURIEs

        # Add alt ids
        alt_ids = self.get_alternative_ids()
        for node in self.nodes:
            node.alternative_ids = alt_ids.get(node.id, [])

        return self

    def get_alternative_ids(self) -> Mapping[str, List[str]]:
        """Get a mapping of primary identifiers to secondary identifiers."""
        rv = defaultdict(set)
        for node in self.nodes:
            if replaced_by := node.replaced_by:
                rv[replaced_by].add(node.id)
        return {k: sorted(v) for k, v in rv.items()}

    def nodes_from(self, prefix: str) -> Iterable[Node]:
        """Iterate non-deprecated nodes whose identifiers start with the given prefix."""
        for node in self.nodes:
            if node.deprecated:
                continue
            if not node.id.startswith(prefix):
                continue
            yield node


def _clean_uri(s: str, *, keep_invalid: bool) -> Optional[str]:
    s = _compress_uri(s)
    n = normalize_curie(s)
    if n:
        return n
    elif keep_invalid:
        return s
    else:
        return None


def _compress_uri(s: str) -> str:
    if s.startswith(OBO_URI_PREFIX):
        s = s[len(OBO_URI_PREFIX) :]
        if s.split("_")[0].isupper():
            s = s.replace("_", ":", 1)
    if s.startswith("http://www.geneontology.org/formats/oboInOwl#"):
        s = s[len("http://www.geneontology.org/formats/oboInOwl#") :]
        s = "oboinowl:" + s
    return s


class GraphDocument(BaseModel):
    """Represents a list of OBO graphs."""

    graphs: List[Graph]
